Package: growthcleanr
Type: Package
Title: Data Cleaner for Anthropometric Measurements
Version: 2.2.3
Authors@R: c(
    person("Carrie", "Daymont", email = "cdaymont@pennstatehealth.psu.edu", role = c("ctb","cre")),
    person("Robert", "Grundmeier", role = "aut"),
    person("Jeffrey", "Miller", role = "aut"),
    person("Diego", "Campos", role = "aut"),
    person("Dan", "Chudnov", email = "dlchudnov@mitre.org", role = "ctb"),
    person("Hannah", "De los Santos", email = "hdelossantos@mitre.org", role = c("ctb")),
    person("Lusha", "Cao", role = "ctb"),
    person("Steffani", "Silva", role = "ctb"),
    person("Hanzhe", "Zhang", role = "ctb"),
    person("Matt", "Boyas", role = "ctb"),
    person("David", "Freedman", role = "ctb"),
    person("Andreas", "Achilleos", role = "ctb"),
    person("Jessica", "Butts", role = "ctb"),
    person("Sheila", "Nguyen", role = "ctb"),
    person("Taraneh", "Soleymani", role = "ctb"),
    person("Max", "Olivier", role = "ctb")
  )
Description: Identifies implausible anthropometric (e.g., height,
    weight) measurements in irregularly spaced longitudinal datasets, such as those from electronic health records.
URL: https://carriedaymont.github.io/growthcleanr/index.html,
        https://github.com/carriedaymont/growthcleanr
BugReports: https://github.com/carriedaymont/growthcleanr/issues
Imports: R.utils (>= 2.11.0), data.table (>= 1.13.0), tidyr (>= 1.1.0),
        plyr (>= 1.8.6), dplyr (>= 1.0.1), foreach (>= 1.5.0),
        doParallel (>= 1.0.15), labelled (>= 2.5.0), magrittr (>= 1.5),
        rlang (>= 0.4.0)
Depends: R (>= 2.10)
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
Suggests: argparser (>= 0.6), bit64 (>= 4.0.2), knitr (>= 1.29),
        rmarkdown (>= 2.3), testthat (>= 2.3.2)
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-09 14:44:19 UTC; carriedaymont
Author: Carrie Daymont [ctb, cre],
  Robert Grundmeier [aut],
  Jeffrey Miller [aut],
  Diego Campos [aut],
  Dan Chudnov [ctb],
  Hannah De los Santos [ctb],
  Lusha Cao [ctb],
  Steffani Silva [ctb],
  Hanzhe Zhang [ctb],
  Matt Boyas [ctb],
  David Freedman [ctb],
  Andreas Achilleos [ctb],
  Jessica Butts [ctb],
  Sheila Nguyen [ctb],
  Taraneh Soleymani [ctb],
  Max Olivier [ctb]
Maintainer: Carrie Daymont <cdaymont@pennstatehealth.psu.edu>
Repository: CRAN
Date/Publication: 2026-05-09 15:20:02 UTC
Built: R 4.6.0; ; 2026-05-09 18:00:08 UTC; unix
