# Generated by roxygen2: do not edit by hand

export(acf_answers)
export(adjustcarryforward)
export(cleangrowth)
export(ewma)
export(ext_bmiz)
export(longwide)
export(read_anthro)
export(recode_sex)
export(sd_median)
export(simple_bmi)
export(splitinput)
export(testacf)
import(R.utils, except = c(extract))
import(data.table)
import(doParallel)
import(dplyr, except = c(last, first, summarize, src, between))
import(foreach)
import(labelled)
import(magrittr)
import(parallel)
import(plyr, except = c(failwith, id, summarize, count, desc, mutate, arrange, rename, is.discrete, summarise, summarize))
import(tidyr, except = extract)
importFrom(rlang,.data)
importFrom(stats,approx)
importFrom(stats,embed)
importFrom(stats,median)
importFrom(stats,pnorm)
importFrom(stats,qnorm)
importFrom(stats,runif)
importFrom(stats,setNames)
importFrom(utils,read.csv)
importFrom(utils,read.table)
importFrom(utils,tail)
importFrom(utils,write.csv)
