#'
NULL

#' Example electronic health record (heightcm, weightkg) data.
#'
#' @title syngrowth
#' @description A synthetic set of measurements from ~3,500
#'   subjects generated using Synthea, with measurement errors
#'   for testing with growthcleanr. Contains both pediatric and adult data.
#'
#' @format A data frame with six variables: \code{id}, \code{subjid},
#'   \code{sex}, \code{agedays}, \code{param}, and \code{measurement}
"syngrowth"
