% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extdata.R
\name{CDCref_d}
\alias{CDCref_d}
\title{CDC BMI reference data}
\description{
Used for extended BMIz computation
}
\section{CDCref_d.csv.gz}{


Used for extended BMI computation
}

